/*
 * Copyright (C) 2006 Apple Computer, Inc.
 * Copyright (C) 2006 Michael Emmel mike.emmel@gmail.com
 * Copyright (C) 2007, 2008 Alp Toker <alp@atoker.com>
 * Copyright (C) 2007 Holger Hans Peter Freyther
 * Copyright (C) 2009, 2010 Igalia S.L.
 * Copyright (c) 2012 NINTENDO.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "FontPlatformData.h"
#if PLATFORM(WIIU) || PLATFORM(WIIU_DEV)
#include "NotImplemented.h"
#include <algorithm>
#include "GraphicsTypes.h"
#include "GraphicsContextPlatformPrivate.h"
#include <wkgfx_Misc.h>
#endif

#include "PlatformString.h"
#include "FontDescription.h"
#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
#include <cairo-ft.h>
#include <cairo.h>
#include <fontconfig/fcfreetype.h>
#else
#include "StatisticsWiiU.h"
#endif

#if !PLATFORM(EFL) && !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
#include <gdk/gdk.h>
#endif

#if !PLATFORM(WIIU)
typedef unsigned char u8;
typedef int s32;
typedef unsigned int u32;
typedef float f32;
#endif

using namespace WTF;
namespace WebCore {

#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)

cairo_subpixel_order_t convertFontConfigSubpixelOrder(int fontConfigOrder)
{
    switch (fontConfigOrder) {
    case FC_RGBA_RGB:
        return CAIRO_SUBPIXEL_ORDER_RGB;
    case FC_RGBA_BGR:
        return CAIRO_SUBPIXEL_ORDER_BGR;
    case FC_RGBA_VRGB:
        return CAIRO_SUBPIXEL_ORDER_VRGB;
    case FC_RGBA_VBGR:
        return CAIRO_SUBPIXEL_ORDER_VBGR;
    case FC_RGBA_NONE:
    case FC_RGBA_UNKNOWN:
        return CAIRO_SUBPIXEL_ORDER_DEFAULT;
    }
    return CAIRO_SUBPIXEL_ORDER_DEFAULT;
}

cairo_hint_style_t convertFontConfigHintStyle(int fontConfigStyle)
{
    switch (fontConfigStyle) {
    case FC_HINT_NONE:
        return CAIRO_HINT_STYLE_NONE;
    case FC_HINT_SLIGHT:
        return CAIRO_HINT_STYLE_SLIGHT;
    case FC_HINT_MEDIUM:
        return CAIRO_HINT_STYLE_MEDIUM;
    case FC_HINT_FULL:
        return CAIRO_HINT_STYLE_FULL;
    }
    return CAIRO_HINT_STYLE_NONE;
}

void setCairoFontOptionsFromFontConfigPattern(cairo_font_options_t* options, FcPattern* pattern)
{
    FcBool booleanResult;
    int integerResult;

    if (FcPatternGetInteger(pattern, FC_RGBA, 0, &integerResult) == FcResultMatch) {
        cairo_font_options_set_subpixel_order(options, convertFontConfigSubpixelOrder(integerResult));

        // Based on the logic in cairo-ft-font.c in the cairo source, a font with
        // a subpixel order implies that is uses subpixel antialiasing.
        if (integerResult != FC_RGBA_NONE)
            cairo_font_options_set_antialias(options, CAIRO_ANTIALIAS_SUBPIXEL);
    }

    if (FcPatternGetBool(pattern, FC_ANTIALIAS, 0, &booleanResult) == FcResultMatch) {
        // Only override the anti-aliasing setting if was previously turned off. Otherwise
        // we'll override the preference which decides between gray anti-aliasing and
        // subpixel anti-aliasing.
        if (!booleanResult)
            cairo_font_options_set_antialias(options, CAIRO_ANTIALIAS_NONE);
        else if (cairo_font_options_get_antialias(options) == CAIRO_ANTIALIAS_NONE)
            cairo_font_options_set_antialias(options, CAIRO_ANTIALIAS_GRAY);
    }

    if (FcPatternGetInteger(pattern, FC_HINT_STYLE, 0, &integerResult) == FcResultMatch)
        cairo_font_options_set_hint_style(options, convertFontConfigHintStyle(integerResult));
    if (FcPatternGetBool(pattern, FC_HINTING, 0, &booleanResult) == FcResultMatch && !booleanResult)
        cairo_font_options_set_hint_style(options, CAIRO_HINT_STYLE_NONE);
}

static cairo_font_options_t* getDefaultFontOptions()
{
#if PLATFORM(GTK)
    if (GdkScreen* screen = gdk_screen_get_default()) {
        const cairo_font_options_t* screenOptions = gdk_screen_get_font_options(screen);
        if (screenOptions)
            return cairo_font_options_copy(screenOptions);
    }
#endif
    return cairo_font_options_create();
}

#endif

#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
FontPlatformData::FontPlatformData(FcPattern* pattern, const FontDescription& fontDescription)
    : m_pattern(pattern)
    , m_fallbacks(0)
    , m_size(fontDescription.computedPixelSize())
    , m_syntheticBold(false)
    , m_syntheticOblique(false)
    , m_fixedWidth(false)
    , m_scaledFont(0)
{

	RefPtr<cairo_font_face_t> fontFace = adoptRef(cairo_ft_font_face_create_for_pattern(m_pattern.get()));
    initializeWithFontFace(fontFace.get());

    int spacing;
    if (FcPatternGetInteger(pattern, FC_SPACING, 0, &spacing) == FcResultMatch && spacing == FC_MONO)
        m_fixedWidth = true;

    if (fontDescription.weight() >= FontWeightBold) {
        // The FC_EMBOLDEN property instructs us to fake the boldness of the font.
        FcBool fontConfigEmbolden;
        if (FcPatternGetBool(pattern, FC_EMBOLDEN, 0, &fontConfigEmbolden) == FcResultMatch)
            m_syntheticBold = fontConfigEmbolden;
    }
}

#else
FontPlatformData::FontPlatformData(struct font_config_handle *handle, font_param *pattern, const FontDescription& fontDescription)
    : m_size(fontDescription.computedPixelSize())
    , m_syntheticBold(false)
    , m_syntheticOblique(false)
    , m_fixedWidth(false)
    , m_scaledFont(0)
{
	m_handle = handle;

	m_pattern.family_name = pattern->family_name;
	m_pattern.style_flags = pattern->style_flags;
	m_pattern.characters = pattern->characters;
	m_pattern.num_chars = pattern->num_chars;
	m_pattern.font_weight = pattern->font_weight;

    m_scaledFont = font_face_t::create();
} 
#endif

FontPlatformData::FontPlatformData(float size, bool bold, bool italic)
    : m_size(size)
    , m_syntheticBold(bold)
    , m_syntheticOblique(italic)
    , m_fixedWidth(false)
    , m_scaledFont(0)
{ 
#if PLATFORM(WIIU) || PLATFORM(WIIU_DEV)
    m_scaledFont = NULL;
    m_pattern.family_name = NULL;
    m_pattern.style_flags = 0;
    m_pattern.characters = NULL;
    m_pattern.num_chars = 0;
    m_pattern.font_weight = 0;
#else
    m_fallbacks = 0;
#endif
    m_handle = NULL;
}

#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
FontPlatformData::FontPlatformData(cairo_font_face_t* fontFace, float size, bool bold, bool italic)
    : m_fallbacks(0)
    , m_size(size)
    , m_syntheticBold(bold)
    , m_syntheticOblique(italic)
    , m_fixedWidth(false)
    , m_scaledFont(0)
{
    initializeWithFontFace(fontFace);

    FT_Face fontConfigFace = cairo_ft_scaled_font_lock_face(m_scaledFont);
    if (fontConfigFace) {
        m_fixedWidth = fontConfigFace->face_flags & FT_FACE_FLAG_FIXED_WIDTH;
        cairo_ft_scaled_font_unlock_face(m_scaledFont);
    }
}
#else
FontPlatformData::FontPlatformData(font_face_t *fontFace, float size, bool bold, bool italic)
    : m_size(size)
    , m_syntheticBold(bold)
    , m_syntheticOblique(italic)
    , m_fixedWidth(false)
    , m_scaledFont(fontFace)
{
	if (m_scaledFont)
		m_scaledFont->use();
	m_handle = NULL;
}

#endif

FontPlatformData& FontPlatformData::operator=(const FontPlatformData& other)
{
    // Check for self-assignment.
    if (this == &other)
        return *this;

    m_size = other.m_size;
    m_syntheticBold = other.m_syntheticBold;
    m_syntheticOblique = other.m_syntheticOblique;
    m_fixedWidth = other.m_fixedWidth;

#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
    if (m_fallbacks) {
        FcFontSetDestroy(m_fallbacks);
        // This will be re-created on demand.
        m_fallbacks = 0;
    }

    if (m_scaledFont && m_scaledFont != hashTableDeletedFontValue())
        cairo_scaled_font_destroy(m_scaledFont);
    m_scaledFont = cairo_scaled_font_reference(other.m_scaledFont);
#else
    
	m_handle = other.m_handle;

    m_pattern.family_name = other.m_pattern.family_name;
    m_pattern.style_flags = other.m_pattern.style_flags;
    m_pattern.characters = other.m_pattern.characters;
    m_pattern.num_chars = other.m_pattern.num_chars;
    m_pattern.font_weight = other.m_pattern.font_weight;
	
	if (m_scaledFont)
		m_scaledFont->destroy();
    m_scaledFont = other.m_scaledFont;
	if (m_scaledFont)
		m_scaledFont->use();
#endif

    return *this;
}

FontPlatformData::FontPlatformData(const FontPlatformData& other)
    : m_scaledFont(0)
{
#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
    m_fallbacks = 0;
#endif
    *this = other;
}

FontPlatformData::FontPlatformData(const FontPlatformData& other, float size)
	: m_scaledFont(0)
{
    // We need to reinitialize the instance, because the difference in size 
    // necessitates a new scaled font instance.
#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
    *this = other;

    m_size = size;
    initializeWithFontFace(cairo_scaled_font_get_font_face(m_scaledFont));
#else
    m_size = size;

	// Custom copy code, to avoid useless interaction with m_scaledFont from 'other'
	m_syntheticBold = other.m_syntheticBold;
	m_syntheticOblique = other.m_syntheticOblique;
	m_fixedWidth = other.m_fixedWidth;

	m_handle = other.m_handle;
	
	m_pattern.family_name = other.m_pattern.family_name;
	m_pattern.style_flags = other.m_pattern.style_flags;
	m_pattern.characters = other.m_pattern.characters;
	m_pattern.num_chars = other.m_pattern.num_chars;
    m_pattern.font_weight = other.m_pattern.font_weight;
    m_scaledFont = font_face_t::create();
#endif
}

FontPlatformData::~FontPlatformData()
{
#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
    if (m_fallbacks) {
        FcFontSetDestroy(m_fallbacks);
        m_fallbacks = 0;
    }

    if (m_scaledFont && m_scaledFont != hashTableDeletedFontValue())
        cairo_scaled_font_destroy(m_scaledFont);
#else
	if (m_scaledFont) {
		m_scaledFont->destroy();
		STATS_WIIU_DEC_FONTCACHE_COUNT();
	}
#endif
}

bool FontPlatformData::isFixedPitch()
{
    return m_fixedWidth;
}

bool FontPlatformData::operator==(const FontPlatformData& other) const
{
#if PLATFORM(WIIU) || PLATFORM(WIIU_DEV)

    if (isHashTableDeletedValue() != other.isHashTableDeletedValue()) {
        return false;
    }

	bool pattern_result = false;
	if (m_pattern.family_name == NULL && other.m_pattern.family_name == NULL) {
		pattern_result = true;
	} else if (m_pattern.family_name == hashTableDeletedFontValue() && other.m_pattern.family_name == hashTableDeletedFontValue()) {
		pattern_result = true;
	} else if (m_pattern.family_name && m_pattern.family_name != hashTableDeletedFontValue() && other.m_pattern.family_name && other.m_pattern.family_name != hashTableDeletedFontValue()) {
		pattern_result = (strcmp(m_pattern.family_name, other.m_pattern.family_name) == 0)
            && m_pattern.font_weight == other.m_pattern.font_weight
            && m_pattern.style_flags == other.m_pattern.style_flags;
	}

    if (!pattern_result) {
        return false;
    }

    bool result = m_size == other.m_size
    	     && m_syntheticOblique == other.m_syntheticOblique
    	     && m_syntheticBold == other.m_syntheticBold;

#else
    bool result = m_pattern == other.m_pattern
        && m_scaledFont == other.m_scaledFont
        && m_size == other.m_size
        && m_syntheticOblique == other.m_syntheticOblique
        && m_syntheticBold == other.m_syntheticBold; 

#endif
    return result; 
}

#ifndef NDEBUG
String FontPlatformData::description() const
{
    return String();
}
#endif

#if !PLATFORM(WIIU) && !PLATFORM(WIIU_DEV)
void FontPlatformData::initializeWithFontFace(cairo_font_face_t* fontFace)
{
    cairo_font_options_t* options = getDefaultFontOptions();

    cairo_matrix_t ctm;
    cairo_matrix_init_identity(&ctm);

    // Scaling a font with width zero size leads to a failed cairo_scaled_font_t instantiations.
    // Instead we scale we scale the font to a very tiny size and just abort rendering later on.
    float realSize = m_size ? m_size : 1;

    cairo_matrix_t fontMatrix;
    if (!m_pattern)
        cairo_matrix_init_scale(&fontMatrix, realSize, realSize);
    else {
        setCairoFontOptionsFromFontConfigPattern(options, m_pattern.get());

        // FontConfig may return a list of transformation matrices with the pattern, for instance,
        // for fonts that are oblique. We use that to initialize the cairo font matrix.
        FcMatrix fontConfigMatrix, *tempFontConfigMatrix;
        FcMatrixInit(&fontConfigMatrix);

        // These matrices may be stacked in the pattern, so it's our job to get them all and multiply them.
        for (int i = 0; FcPatternGetMatrix(m_pattern.get(), FC_MATRIX, i, &tempFontConfigMatrix) == FcResultMatch; i++)
            FcMatrixMultiply(&fontConfigMatrix, &fontConfigMatrix, tempFontConfigMatrix);
        cairo_matrix_init(&fontMatrix, fontConfigMatrix.xx, -fontConfigMatrix.yx,
                          -fontConfigMatrix.xy, fontConfigMatrix.yy, 0, 0);

        // The matrix from FontConfig does not include the scale. 
        cairo_matrix_scale(&fontMatrix, realSize, realSize);
    }

    m_scaledFont = cairo_scaled_font_create(fontFace, &fontMatrix, &ctm, options);
    cairo_font_options_destroy(options);
}
#endif

} // namespace WebCore

