/*
 * Copyright (C) 2010 Google Inc. All rights reserved.
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#if USE(ACCELERATED_COMPOSITING)

#include "ContentLayerWiiU.h"
#include "ImageBuffer.h" 
#include "GraphicsContextPlatformPrivate.h"
#include "wtf/text/CString.h"
// Other implementations break large images into tiles and then render the tiles individually.
// We have a "tiling size" which is the largest texture we will generate. Since we already have
// an accelerated renderer, if the image is too large, we just fallback to our standard rendering
// path, rather than tiling. We may want to revisit this decision in the future, time permiting, to
// see if tiling will gain us performance relative to using the standard renderer.
//
// We use 2048 rathern 8192 (maximum hardware dimension) so that we don't chew up a bunch of memory
// for dubious performance gains.
//
const int TILING_SIZE = 2048;

namespace WebCore {

PassRefPtr<ContentLayerWiiU> ContentLayerWiiU::create(GraphicsLayerWiiU* owner)
{
    return adoptRef(new ContentLayerWiiU(owner));
}

ContentLayerWiiU::ContentLayerWiiU(GraphicsLayerWiiU* owner)
    : LayerWiiU(owner)
{
}

ContentLayerWiiU::~ContentLayerWiiU()
{
    cleanupResources();
}

void ContentLayerWiiU::cleanupResources()
{
    if (m_buffer) {
        m_buffer.release();
    } 
}

void ContentLayerWiiU::paintContentLayerIfDirty()
{
    int width  = contentBounds().width();
    int height = contentBounds().height();
    //OS::Report("Paint dirty Layer %s Pos (%fx%f) Size (%dx%d)\n", name().utf8().data(), m_owner->position().x(), m_owner->position().y(), width, height);

    if (drawsContent()  && m_contentsDirty && (width > 0) && (width < TILING_SIZE) && (height > 0) && (height < TILING_SIZE)) {
        int count = m_owner->incrementRepaintCount();
        //see if we can re-use our current buffer
        if (m_buffer) {
            if (m_buffer->width() != width || m_buffer->height() != height) {
                m_buffer.release();
                m_buffer = ImageBuffer::create((IntSize(width, height)));
            }
        } else {
            m_buffer = ImageBuffer::create((IntSize(width, height)));
        }

        //see if we got an image buffer, it is possible that the allocation could fail
        //if that happens, we'll drop through to the "slow" render path in the paint
        //call.
        if (m_buffer) {
            m_buffer->context()->save();
            m_buffer->context()->clearRect(FloatRect(0.0, 0.0, (float)width, (float)height));
            m_owner->paintGraphicsLayerContents(*m_buffer->context(), IntRect(0, 0, width, height));
            m_buffer->context()->restore();
        }

        m_contentsDirty = false;
        m_dirtyRect = FloatRect();

        if (m_owner->showDebugBorders()) {
            drawDebugBorder(m_buffer->context(), width, height);
        }
    } 
}

void ContentLayerWiiU::paintContents(GraphicsContext *ctx)
{
    ASSERT(ctx);

    int width  = contentBounds().width();
    int height = contentBounds().height();

    if (m_buffer) {    
         ctx->drawImageBuffer(m_buffer.get(),ColorSpaceDeviceRGB,  IntRect(0, 0, width, height), IntRect(0, 0, m_buffer->width(), m_buffer->height()));
    } else {
         m_owner->paintGraphicsLayerContents(*ctx, IntRect(0, 0, width, height));
    }
}

bool ContentLayerWiiU::drawsContent() const
{
    return m_owner && m_owner->drawsContent();
}

}
#endif // USE(ACCELERATED_COMPOSITING)
