/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*

	Video Playback for WiiU
	(high level interface with Webkit)

*/
#include "config.h"

#if ENABLE(VIDEO) && (PLATFORM(WIIU))

// --------------------------------------------------------------------

#include "MediaPlayerPrivateWiiU.h"
#include "TimeRanges.h"
#include "DomainAccessControlWiiU.h"
#include "GraphicsContext.h"
#include "GraphicsContextPlatformPrivate.h"
#include "Frame.h"
#include "FrameLoaderClientWiiU.h"

using namespace std;

String	WebCore::MediaPlayerPrivateWiiU::s_lastQueryType;
String	WebCore::MediaPlayerPrivateWiiU::s_lastQueryCodecs;

namespace WebCore {

// --------------------------------------------------------------------

// --------------------------------------------------------------------
/**
	Support function to not have to rebuild the MIME info over and over again
*/
static HashSet<String> mimeTypeCache()
	{
	DEFINE_STATIC_LOCAL(HashSet<String>, cache, ());
	static bool _bTypeListInitialized = false;

	if(!_bTypeListInitialized)
		{
		cache.add(String("audio/ogg"));						// Ogg Vorbis
// MIME types below are now handled by MediaPlayerPrivateMobiclip
#ifndef ENABLE_MOBICLIP_MP	

#ifdef ENABLE_MPEG_DECODER
		cache.add(String("audio/mpeg"));					// Mpeg audio, layers 1/2/3
#endif		
		cache.add(String("application/vnd.apple.mpegurl"));	// HLS
		cache.add(String("vnd.apple.mpegurl"));
		cache.add(String("application/x-mpegurl"));			// HLS, too
		cache.add(String("audio/aac"));						// AAC
		cache.add(String("video/mp4"));						// H.264 & possibly AAC
		cache.add(String("video/mp2t"));					// MP2TS
#endif
		_bTypeListInitialized = true;
		}
	return cache;
	}

// --------------------------------------------------------------------
/**
	Register the supported MIME types of this media engine
*/
void MediaPlayerPrivateWiiU::getSupportedTypes(HashSet<String>& types)
{
	const HashSet<String>& mimeCache = mimeTypeCache();

	HashSet<String>::const_iterator it = mimeCache.begin();
    HashSet<String>::const_iterator end = mimeCache.end();
    for (; it != end; ++it )
            types.add( *it );
}

// -------------------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------------------

// --------------------------------------------------------------------
/**
	Create an instance of this media engine
*/
PassOwnPtr<MediaPlayerPrivateInterface> MediaPlayerPrivateWiiU::create(MediaPlayer* player)
{
    return adoptPtr(new MediaPlayerPrivateWiiU(player));
}

// --------------------------------------------------------------------
/**
	Check if this media engine is available
*/
bool MediaPlayerPrivateWiiU::isAvailable()
{
	// In this context we assume always available player
	return true;
}

// --------------------------------------------------------------------
/**
	Check if we support a given type
*/
MediaPlayer::SupportsType MediaPlayerPrivateWiiU::supportsType(const String& type, const String& codecs)
{
	s_lastQueryType = type;
	s_lastQueryCodecs = codecs;

	if (type.isNull() && type.isEmpty())
		return MediaPlayer::IsNotSupported;

    // spec says we should not return "probably" if the codecs string is empty
    if (mimeTypeCache().contains(type))
        return codecs.isEmpty() ? MediaPlayer::MayBeSupported : MediaPlayer::IsSupported;
    return MediaPlayer::IsNotSupported;
}

// --------------------------------------------------------------------
/**
	Register the media engine with Webkit
*/
void MediaPlayerPrivateWiiU::registerMediaEngine(MediaEngineRegistrar registrar)
{
	if (isAvailable())
        registrar(create, getSupportedTypes, supportsType, 0, 0, 0);
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::doLoad(const String& url)
{
	m_url = encodeWithURLEscapeSequences(url).latin1().data();

	m_type = s_lastQueryType.utf8().data();
	m_codecs = s_lastQueryCodecs.utf8().data();

OS::Report("Load %s MIME=%s CODECS=%s\n", m_url.c_str(), m_type.c_str(), m_codecs.c_str());

	if (m_preLoadMode == MediaPlayer::Auto)
		openStream();
}

// --------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateWiiU::isURLallowable(const std::string & url)
	{
			int error = WebCore::DomainAccessControl::DOMAIN_ACCESS_OK;
	if (!WebCore::DomainAccessControl::IsDomainAccessible(url.c_str(), &error))
			{
				switch (error)
			{
				case WebCore::DomainAccessControl::DOMAIN_ACCESS_NOT_WHITELISTED:
					OS::Report("The url '%s' is not whitelisted. Add its domain to config.xml to whitelist.\n", url.c_str());
					break;

				case WebCore::DomainAccessControl::DOMAIN_ACCESS_NOT_ACQUIRED_AUTH_TOKEN:
					OS::Report("Before accessing a network resource, an independent service token must first be acquired.\n" \
							"Please call nsf.act.NintendoAccountManager.acquireNewIndependentServiceToken().\n");
					break;

				default:
					OS::Report("This domain is unreachable. Error %d\n", error);
					break;
				}
				return false;
			}
	return true;
}

// --------------------------------------------------------------------------
/**
	Return index of display the player is on

	@return	Index of display (TV=0, GAMEPAD=1,2...)
*/

int MediaPlayerPrivateWiiU::getDisplayIndex() const
{
	WebView *pView = static_cast<FrameLoaderClientWiiU*>(m_player->mediaPlayerClient()->mediaPlayerOwningDocument()->frame()->loader()->client())->webView();

	if (pView->GetDisplayType() == WebView::GAMEPAD_DISPLAY_TYPE)
		return 1 + pView->GetGamePadIndex();
	return 0;
}

// --------------------------------------------------------------------------
/**
*/
Timer<MediaTimer>* MediaPlayerPrivateWiiU::CreateTimer(MediaTimer *pTimer)
{
	return new Timer<MediaTimer>(pTimer, &MediaTimer::_timerCallback);
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::DestroyTimer(Timer<MediaTimer>* pTimer)
{
	delete pTimer;
}

// --------------------------------------------------------------------------
/**
*/
void  MediaPlayerPrivateWiiU::StartTimer(Timer<MediaTimer>* pTimer, double period)
{
	pTimer->startRepeating(period);
}

// --------------------------------------------------------------------------
/**
*/
void  MediaPlayerPrivateWiiU::StopTimer(Timer<MediaTimer>* pTimer)
{
	pTimer->stop();
}

// --------------------------------------------------------------------------
/**
 */
void MediaTimer::_timerCallback(WebCore::Timer<MediaTimer>*)
{
	timerCallback();
}

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// - Interface for wksupport calls ------------------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

// --------------------------------------------------------------------------
/**
*/
Timer<MediaTimer>* GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayerPrivateWiiU_CreateTimer(MediaTimer *pTimer)
{
	return MediaPlayerPrivateWiiU::CreateTimer(pTimer);
}

// --------------------------------------------------------------------------
/**
*/
void GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayerPrivateWiiU_DestroyTimer(Timer<MediaTimer>* pTimer)                                                                
{
	MediaPlayerPrivateWiiU::DestroyTimer(pTimer);
}

// --------------------------------------------------------------------------
/**
*/
void GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayerPrivateWiiU_StartTimer(Timer<MediaTimer>* pTimer, double period)                                                      
{
	MediaPlayerPrivateWiiU::StartTimer(pTimer, period);
}

// --------------------------------------------------------------------------
/**
*/
void GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayerPrivateWiiU_StopTimer(Timer<MediaTimer>* pTimer)                                                                   
{
	MediaPlayerPrivateWiiU::StopTimer(pTimer);
}

// --------------------------------------------------------------------------
/**
*/
int GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayerPrivateWiiU_getDisplayIndex(const MediaPlayerPrivateWiiU *pThis)                                                    
{
	return pThis->getDisplayIndex();
}

// --------------------------------------------------------------------------
/**
*/
bool GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayerPrivateWiiU_isURLallowable(const std::string & url)                                                                         
{
	return MediaPlayerPrivateWiiU::isURLallowable(url);
	}

// --------------------------------------------------------------------------
/**
*/
void GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayerPrivateWiiU_doLoad(MediaPlayerPrivateWiiU *pThis, const WTF::String & url)                                           
{
	pThis->doLoad(url);
}

// --------------------------------------------------------------------------
/**
*/
void GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayer_networkStateChanged(MediaPlayer *pThis)                                                                           
{
	pThis->networkStateChanged();
}

// --------------------------------------------------------------------------
/**
*/
void GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayer_readyStateChanged(MediaPlayer *pThis)                                                                             
{
	pThis->readyStateChanged();
}

// --------------------------------------------------------------------------
/**
*/
void GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayer_timeChanged(MediaPlayer *pThis)                                                                                   
	{
	pThis->timeChanged();
}

// --------------------------------------------------------------------------
/**
*/
void GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayer_repaint(MediaPlayer *pThis)                                                                                       
{
	pThis->repaint();
}

// --------------------------------------------------------------------------
/**
*/
void GraphicsContextPlatformPrivate::WKgfxHooks::GraphicsContext_fillRect(GraphicsContext *pThis, const FloatRect& r, const Color& c, ColorSpace cs)
{
	pThis->fillRect(r, c, cs);
}

// --------------------------------------------------------------------------
/**
*/
bool GraphicsContextPlatformPrivate::WKgfxHooks::GraphicsContext_paintingDisabled(const GraphicsContext *pThis)                                                                
{
	return pThis->paintingDisabled();
}

// --------------------------------------------------------------------------
/**
*/
const PlatformMedia & GraphicsContextPlatformPrivate::WKgfxHooks::getNoPlatformMedia()
{
	return getNoPlatformMedia();
}

}; // namespace WebCore


#endif

