/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __WKSUP_FS_H__
#define __WKSUP_FS_H__

#include <stdlib.h>

#include <types.h>

#if !defined(LGPL)

#endif

#ifdef __cplusplus
extern "C" {
#endif

/// @addtogroup wksupFileSystem
/// @{

typedef s32 wksupFSFileInfo;
typedef s32 wksupFSDir;
typedef s32 wksupFSDirEntry;

// At the moment, only 3 of these result values are ever returned:
// (This may change in the future.)

typedef enum
{
    WKSUP_FS_RESULT_OK              =  0,  // YES!
    WKSUP_FS_RESULT_ACCESS          = -1,  // no
    WKSUP_FS_RESULT_PERMISSION      = -2,  // no
    WKSUP_FS_RESULT_CORRUPTION      = -3,  // no
    WKSUP_FS_RESULT_AREA_FULL       = -4,  // no
    WKSUP_FS_RESULT_MAX_HANDLES     = -5,  // no
    WKSUP_FS_RESULT_ALREADY_EXISTS  = -6,  // no
    WKSUP_FS_RESULT_NOT_FOUND       = -7,  // YES!
    WKSUP_FS_RESULT_NOT_EMPTY       = -8,  // no
    WKSUP_FS_RESULT_FILE_OPEN       = -9,  // no
    WKSUP_FS_RESULT_FATAL_ERROR     = -128 // YES!
} wksupFSResultCode;

#define wksup_INLINE static inline

#define wksupAssert(expr) ASSERT(expr)
#define wksupRoundUp256B(x) (((u32)(x) + 256 - 1) & ~(256 - 1))
#define wksupRoundUp32B(x) (((u32)(x) + 32 - 1) & ~(32 - 1))
#define wksupRoundUp4B(x) (((u32)(x) + 4 - 1) & ~(4 - 1))

/// \brief Reads a file into an automatically created and allocated buffer.  
///
/// This buffer is not aligned in any special way.  
/// File is automatically closed after it's read from.
/// Call \ref wksupFree() to free this buffer.
///
/// \param filename File string name to read
/// \param len Length of file returned here
///
void * wksupFSSimpleRead(const char * filename, u32 * len);

/// \brief Reads a file into an automatically created and allocated buffer with alignment.  
///
/// This buffer is aligned with alignSize.  
/// File is automatically closed after it's read from.
/// Call \ref wksupFree() to free this buffer.
///
/// \param filename File string name to read
/// \param len Length of file returned here
/// \param alignSize Alignment size for buffer
///
void * wksupFSSimpleReadAlign(const char * filename, u32 * len, u32 alignSize);


/// \brief Frees the memory allocated by wksupFSSimpleRead or wksupFSSimpleReadAlign.
///
/// \param addr The buffer to free
///
void wksupFSFree(void* addr);

/// @}

#ifdef __cplusplus
}
#endif

#endif /// __WKSUP_FS_H__
